from marshmallow import Schema

__all__ = ["Serializer"]


class Serializer(Schema):
    async def create(self, data):
        pass

    async def update(self, obj, data):
        pass

    async def get(self, id):
        pass

    async def list(self, filters, page=0, page_size=20, sort=(None, -1)):
        pass

    async def count(self, filters):
        pass

    async def delete(self, id):
        pass

    async def to_representation(self, obj):
        pass



from json import JSONDecodeError
from .exceptions import APIInvalidJSONError, APIContentTypeError


async def cors_signal(request, response):
    """Add some CORS headers"""
    response.headers["Access-Control-Allow-Origin"] = '*'
    response.headers["Access-Control-Allow-Headers"] = 'Content-Type, Authorization'
    response.headers["Access-Control-Allow-Methods"] = 'POST, OPTIONS, GET, PUT, DELETE'


from puer.module import AbstractModule
from aiohttp_swagger import setup_swagger

class Swagger(AbstractModule):
    name = None

    def __init__(self, manager, app):
        settings = manager.settings.swagger
        setup_swagger(app,
                      swagger_url=settings.url,
                      swagger_from_file=settings.yaml)



from json import dumps
from aiohttp.web_exceptions import HTTPException


class APIError(HTTPException):
    status_code = None
    empty_body = False

    def __init__(self, errors=None):
        kwargs = dict()
        kwargs["content_type"] = "application/json"
        kwargs["body"] = dumps({
            "errors": errors or []
        })
        super().__init__(**kwargs)


class InvalidAuthorizationError(APIError):
    status_code = 401


class APIInvalidJSONError(APIError):
    status_code = 400

    def __init__(self):
        super().__init__({"request": "Not valid JSON"})


class APIBadRequest(APIError):
    status_code = 400


class APINeedPermissions(APIError):
    status_code = 403


class APINotFound(APIError):
    status_code = 404

    def __init__(self):
        super().__init__({"request": "Document with requested ID doesn't exist"})


class APIContentTypeError(APIError):
    status_code = 400

    def __init__(self):
        super().__init__({"request": "Content-Type must be 'application/json'"})



from marshmallow.exceptions import ValidationError
from marshmallow.fields import Field

import bson


class ObjectId(Field):
    """
    Marshmallow field for :class:`bson.ObjectId`
    """

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return str(value)

    def _deserialize(self, value, attr, data):
        if value is None:
            return None
        if not isinstance(value, (bytes, str, ObjectId)):
            raise ValidationError('Invalid ObjectId.')
        try:
            return bson.ObjectId(value)
        except bson.errors.InvalidId:
            raise ValidationError('Invalid ObjectId.')


from ...exceptions import APINeedPermissions

__all__ = ["PermissionsMixin", "AuthorizedOnly"]


class AbstractPermissionsBackend(object):
    async def check_permissions(self, request):
        pass


class AuthorizedOnly(AbstractPermissionsBackend):
    async def check_permissions(self, request):
        if request["user"] is None:
            raise APINeedPermissions()


class PermissionsMixin(object):
    permissions_backends = (AuthorizedOnly,)

    async def dispatch(self):
        if self.request._method.lower() != 'options':
            for backend in self.permissions_backends:
                await backend().check_permissions(self.request)
        return await super().dispatch()



import asyncio
from math import ceil

from aiohttp import hdrs
from puer_rest.exceptions import APINotFound, APIBadRequest

from puer_rest.views import APIView
from modules.auth.documents import UserExist

__all__ = ["APIViewSet"]


class APIViewSet(APIView):
    filter_set_schema = None

    method_map = {
        "get": {
            None: "list_view",
            "detail": "detail_view"
        },
        "post": {
            None: "create_view",
            "detail": None
        },
        "put": {
            None: None,
            "detail": "update_view"
        },
        "delete": {
            None: None,
            "detail": "delete_view"
        }
    }

    @property
    def _list_filters(self):
        if self.filter_set_schema is not None:

            filters = self.filter_set_schema().load(self.request.query)

            if any(filters.errors):
                raise APIBadRequest({"filters": filters.errors})
            return filters.data
        return {}

    async def list_view(self):
        serializer = self.get_serializer()
        filters = self._list_filters
        # Sorting
        sort_field = self.request.query.get("sort_field", "_id")
        try:
            sort_direction = int(self.request.query.get("sort_direction", -1))
        except ValueError:
            sort_direction = -1
        # Paginator
        page = int(self.request.query.get("page", 0))
        page_size = int(self.request.query.get("page_size", 20))

        object_list = serializer.list(filters, page, page_size, (sort_field, sort_direction))
        objects_count = await serializer.count(filters)
        pages_count = ceil(objects_count / page_size)
        result = []
        async for obj in await object_list:
            result.append(await serializer.to_representation(obj))

        return self.response(
            {
                "result": result,
                "page": page,
                "page_size": page_size,
                "count": objects_count,
                "pages_count": pages_count
            }
        )

    async def detail_view(self):
        serializer = self.get_serializer()
        obj = await serializer.get(self.request["detail_id"])
        if obj is None:
            raise APINotFound
        res = await serializer.to_representation(obj)
        return self.response(res)

    async def create_view(self):
        serializer = self.get_serializer()
        obj = await serializer.create(self.validated_data)
        res = await serializer.to_representation(obj)
        res["_created"] = True
        return self.response(res)

    async def update_view(self):
        serializer = self.get_serializer()
        obj = await serializer.get(self.request["detail_id"])
        if obj is None:
            raise APINotFound
        updated_obj = await serializer.update(obj, self.validated_data)
        res = await serializer.to_representation(updated_obj)
        res["_updated"] = True
        return self.response(res)

    async def delete_view(self):
        serializer = self.get_serializer()
        deleted = await serializer.delete(self.request["detail_id"])
        return self.response(
            {
                "id": self.request["detail_id"],
                "deleted": deleted
            }
        )

    def get_detail_method(self, method_variants):
        return method_variants.get("detail")

    def get_none_method(self, method_variants):
        return method_variants.get(None)

    def get_method(self):
        method = self.request._method.lower()
        detail = True if self.request.get("detail_id") else None
        method_variants = self.method_map.get(method)
        if method_variants:
            if detail:
                return self.get_detail_method(method_variants)
            else:
                return self.get_none_method(method_variants)
        else:
            return self.request._method.lower()

    async def dispatch(self):
        self.request["detail_id"] = self.request.match_info.get('id')
        return await super().dispatch()



from marshmallow import Schema, post_load


__all__ = ["FilterSetSchema"]


class FilterSetSchema(Schema):
    def dotter(self, d, key, dots):
        if isinstance(d, dict):
            for k in d:
                self.dotter(d[k], key + '.' + k if key else k, dots)
        else:
            dots.update({key: d})
        return dots

    @post_load
    def mongo_filter(self, item):
        return self.dotter(item, '', {})


from puer.route import route


def viewset_route(r, func, name=None):
    """
    
    :param r: (str) route
    :param func: (class|async function) handler
    :param name: route name
    :return: 
    """
    route(r, func, name)
    route("%s{id}/" % r, func, name)


import asyncio
from json import JSONDecodeError

from aiohttp import hdrs
from puer.views import View, Response

from ..exceptions import APIInvalidJSONError, APIBadRequest

__all__ = ["APIView"]


class APIView(View):
    """
    Base REST view with check request data by marshmallow schema (see Meta.data_schema)
    """

    serializer = None

    def __init__(self, request):
        super().__init__(request)
        self.raw_data = {}
        self.validated_data = {}

    def _raise_allowed_methods(self):
        allowed_methods = {
            m for m in hdrs.METH_ALL if hasattr(self, m.lower())}
        raise APIBadRequest()

    @asyncio.coroutine
    def __iter__(self):
        # Check for allowed method
        if self.request._method not in hdrs.METH_ALL:
            self._raise_allowed_methods()

        # Dispatch http method
        resp = yield from self.dispatch()
        return resp

    def get_serializer(self):
        return self.serializer(context={"request": self.request})

    def get_method(self):
        return self.request._method.lower()
        
    async def dispatch(self):
        # Check content_type for application/json, get dict from request JSON
        method_name = self.get_method()
        if method_name is None:
            self._raise_allowed_methods()
        method = getattr(self, method_name, None)
        if method is None:
            self._raise_allowed_methods()

        content_type = self.request.headers.get("Content-Type")
        if self.request._method.lower() not in ('get', 'delete', 'options'):
            if content_type == "application/json":
                try:
                    self.raw_data = await self.request.json()
                except JSONDecodeError:
                    raise APIInvalidJSONError()
                # Data validation
                if self.get_serializer() is not None:
                    self.validate_data()
            else:
                raise APIInvalidJSONError()
        return await method()

    def validate_data(self):
        serializer = self.get_serializer()
        marshmallow_result = serializer.load(self.raw_data)
        errors = marshmallow_result.errors
        if any(errors):
            raise APIBadRequest({"fields": errors})
        self.validated_data = marshmallow_result.data

    async def options(self):
        return Response()

        




from marshmallow.fields import *
from .object_id import ObjectId
from umongo.marshmallow_bonus import GenericReference as umGR, Reference

from datetime import datetime, date
import dateutil.parser
from marshmallow import utils


class GenericReference(umGR):
    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        if self.mongo_world:
            # In mongo world, value a dict of cls and id
            return {'id': str(value['_id']), 'cls': value['_cls']}
        else:
            if isinstance(value, dict):
                return {'id': str(value['id']), 'cls': value['cls']}
            # In OO world, value is a :class:`umongo.data_object.Reference`
            return {'id': str(value.pk), 'cls': value.document_cls.__name__}


class DateTime(DateTime):
    def _serialize(self, value, attr, obj):
        if not value:  # Falsy values, e.g. '', None, [] now valid
            return None
        if isinstance(value, str):
            value = dateutil.parser.parse(value)
        return super()._serialize(value, attr, obj)

    def _deserialize(self, value, attr, data):
        if not value:  # Falsy values, e.g. '', None, [] now valid
            return None
        return super()._deserialize(value, attr, data)


class Method(Method):
    def _deserialize(self, value, attr, data):
        if self.deserialize_method_name:
            try:
                method = utils.callable_or_raise(
                    getattr(self.parent, self.deserialize_method_name, None)
                )
                return method(attr, data)
            except AttributeError:
                pass
        return value
